# coding=utf-8

from .CfMetric import CfMetric
#from Tools.ListTools import is_list, is_number_list
#from Tools.NumberTools import is_number
from .BaseMetric import BaseMetric
from .PSI import calc_psi_list
import numpy as np
import pandas as pd
import sys
sys.dont_write_bytecode = True

# 基于pandas的
class Metric:
    def __init__(self, data, label_name=None, score_name=None):
        assert isinstance(data, pd.DataFrame), "support DataFrame only (for now)"
        self.data = data
        self.label_name = label_name if label_name is not None else 'gb_flag'
        assert self._check_label_valid(), self.label_name + " should take value from {0, 1}..."
        self.score_name = score_name
        self.len, self.dim = data.shape

    def _check_label_valid(self):
        if self.data[self.label_name].max() == 1 and self.data[self.label_name].min() == 0:
            return True
        return False

    # 返回KS；可选：KS_list和坏账
    def get_ks(self, col_name=None, print_list=False, bad_rate=False):
        col_name = col_name if col_name is not None else self.score_name
        if len(set(self.data[col_name].values)) < 2:
            print("all scores are the same.")
            return -1
        else:
            base_metric = BaseMetric(self.data[col_name].values)
            return base_metric.get_ks(self.get_label_values(), print_list, bad_rate)

    def get_label_values(self):
        return self.data[self.label_name].values

    # KS对比，返回坏账可选
    def compare_ks(self, other_name, my_name=None, other_missing_value=None, my_missing_value=None, bad_rate=False):
        my_name = my_name if my_name is not None else self.score_name
        data_ = self.data.copy()
        if other_missing_value is not None:
            print('remove other')
            data_ =data_[data_[other_name] != other_missing_value]
        if my_missing_value is not None:
            print('remove my')
            data_ =data_[data_[my_name] != my_missing_value]
            
        my_ks = self.get_ks(col_name=my_name, bad_rate=bad_rate)
        other_ks = self.get_ks(col_name=other_name, bad_rate=bad_rate)
        if not bad_rate:
            return my_ks, other_ks
        else:
            return my_ks[0], other_ks

    # =================  分群

    # 逐月PSI（或者其他split上比较）
    def get_psi_per_split(self, split_name='month', col_name=None, bins=10, list_=False, ret_simple=True):
        col_name = col_name if col_name is not None else self.score_name
        months = sorted(set(self.data[split_name].values))
        re = ''
        for ind in range(1, len(months)):
            tmp1 = self.data[self.data[split_name] == months[ind - 1]]
            tmp2 = self.data[self.data[split_name] == months[ind]]
            cs = calc_psi_list(tmp1[col_name].values, tmp2[col_name].values, nbins=bins)
            if ret_simple:
                re += "|" + str(cs["psi_value"].sum())
            else:
                re+= "{}~{}|PSI|{}\n".format(months[ind - 1], months[ind], cs["psi_value"].sum())
            if list_:
                print(cs["psi_value"].values)
        return re

    # 逐月PSI（或者其他split上比较）
    def get_psi_given_splits(self, splits, split_name='month', col_name=None, bins=10):
        col_name = col_name if col_name is not None else self.score_name
        act_splits = sorted(set(self.data[split_name].values))
        ret = {}
        for spl in splits:      # 计算每个给定split上的spi
            if spl in act_splits:
                # 实际数据中存在 spl的数据
                idx = act_splits.index(spl)
                if idx > 0:
                    # 存在上一个月的数据（不一定连续），计算spi
                    tmp1 = self.data[self.data[split_name] == act_splits[idx - 1]]
                    tmp2 = self.data[self.data[split_name] == act_splits[idx]]
                    #分桶数取最短板
                    bins = min(tmp1.shape[0],tmp2.shape[0],bins)
                    cs = calc_psi_list(tmp1[col_name].values, tmp2[col_name].values, nbins=bins)
                    ret[spl] = cs["psi_value"].sum()
        return ret

    # 逐月KS（或者其他split上比较）
    def get_ks_per_split(self, split_name='month', col_name=None, bad_rate=True):
        col_name = col_name if col_name is not None else self.score_name
        months = sorted(set(self.data[split_name].values))
        results = {}
        for month in months:
            tmp1 = self.data[self.data[split_name] == month]
            base_metric = BaseMetric(tmp1[col_name].values)
            res = base_metric.get_ks(tmp1[self.label_name].values, bad_rate=bad_rate)
            results[month] = res
            print("{} , KS={}".format(month, res))
        return results

    # 交叉分析
    def get_mix(self, other_data,  other_name, bins=10, on_keys=['pin', 'createtime'], my_name=None):
        assert isinstance(other_data, pd.DataFrame), "support DataFrame only (for now)"
        my_name = my_name if my_name is not None else self.score_name
        x = self.data.reset_index()
        ascore_data = other_data.reset_index()
        ascore_data = ascore_data.drop(self.label_name, axis=1)
        merged = pd.merge(x, ascore_data, on=on_keys)
        pred_score = merged[[my_name, self.label_name] + on_keys].sort_values(by=my_name)
        ascore_data = merged[[other_name] + on_keys].sort_values(by=other_name)
        bin_size = int(len(merged) / bins)
        results_bad = np.zeros((bins, bins))
        results_tot = np.zeros((bins, bins))
        result_badrate = np.zeros((bins, bins))
        for i in range(bins):
            print("processing {}th bin".format(i))
            ascore = ascore_data.iloc[i * bin_size: min((i + 1) * bin_size, len(merged))]
            for j in range(bins):
                pred = pred_score.iloc[j * bin_size: min((j + 1) * bin_size, len(merged))]
                merged_small = pd.merge(ascore, pred, on=on_keys)
                results_bad[i][j] = len(merged_small[merged_small[self.label_name] == 0])
                results_tot[i][j] = len(merged_small)
                result_badrate[i][j] = results_bad[i][j] * 1.0 / results_tot[i][j] if results_tot[i][j] > 0 else -1
        return results_tot, results_bad, result_badrate

    # 字段覆盖率
    def get_split_cover(self, col_lists, missing_value=-9999.8):
        # assert is_list(col_lists), "give me list of cols..."
        data_ = self.data.fillna(missing_value)
        for col in col_lists:
            tot_non = len(data_[data_[col] != missing_value])
            print("{}, tot/non={}\t{}\t{}".format(col, self.len, tot_non, tot_non * 1.0 / self.len))

    # 换入换出分析
    def swap_out(self, data, score_threshold, score_name=None, stratigy_name='gb_flag'):
        # assert is_number(score_threshold), "threshold should be a numeric"
        score_name = score_name if score_name is not None else self.score_name
        data[score_name+"_bin"] = data[score_name].apply(lambda x: 1 if x >= score_threshold else 0)
        print("stratigy VS score")
        print(data.groupby([stratigy_name, score_name]).size())

    # lift值和分段坏账
    def get_lift(self, score_name=None, bins=10):
        score_name = score_name if score_name is not None else self.score_name
        y = self.get_label_values()
        score_pred = self.data[score_name].values
        tot_bad = np.sum(y == 0)
        kv = list(zip(score_pred, y))
        kv_sorted = sorted(kv, key=lambda x: x[0])
        label_sorted = list(zip(*kv_sorted))[1]
        if len(label_sorted) % bins != 0:
            gewei = len(label_sorted) % 10
            label_sorted = label_sorted[: -1 * int(gewei)]
        bin_size = float(len(label_sorted) / bins)
        label_sorted = np.reshape(label_sorted, (bins, -1))
        num_goods = np.sum(label_sorted, axis=1)
        num_bads = bin_size - num_goods
        lift = num_bads / tot_bad
        bad_rad = 1 - num_goods / bin_size
        return bad_rad, lift

    # 分数段划分
    def split(self, bin=10, col_name=None, index_=None):
        # assert is_number(bin) or is_number_list(bin), "bin should be numeric or a numeric list"
        if index_ is None:
            index_ = "pin"
        self.data = self.data.set_index(index_)
        col_name = col_name if col_name is not None else self.score_name
        a = pd.qcut(self.data[col_name], bin)
        self.data[col_name+"_bin"] = a
        self.data = self.data.reset_index()
        print(a.value_counts())

    # 分段坏账 lift 第二版
    def print_lift_v2(self, bin=10, col_name=None, index_=None):
        col_name = col_name if col_name is not None else self.score_name
        print("split data..")
        self.split(bin, col_name, index_)
        print(self.data.groupby([col_name+"_bin", self.label_name]).size())

    # 按照有标签的激活用户来分段，然后应用在申请用户上
    # 保证apply_data中的交叉字段和data一致
    def get_mix_v2(self, other_col, apply_data, bin=10, col_name=None, index_=None):
        col_name = col_name if col_name is not None else self.score_name
        assert isinstance(apply_data, pd.DataFrame), "support DataFrame only (for now)"
        # 激活用户上的分段
        print("split data..")
        self.split(bin, col_name, index_)
        self.split(bin, other_col, index_)
        result = self.data.groupby([col_name+"_bin", other_col + "_bin", self.label_name]).size()
        # 申请数据上的分段
        splits_info = list(set(self.data[col_name+"_bin"].values))
        splits_info_other = list(set(self.data[other_col+"_bin"].values))
        apply_data[other_col+"_bin"] = apply_data[other_col].apply(lambda x: self._check_range(x, splits_info_other))
        apply_data[col_name+"_bin"] = apply_data[col_name].apply(lambda x: self._check_range(x, splits_info))
        result_apply = apply_data.groupby([col_name+"_bin", other_col + "_bin"]).size()
        print(result)
        print(result_apply)
        return result, result_apply

    def _check_range(self, score, list_info):
        for info in list_info:
            if float(score) in info:
                return info
        return -1

    def _count_bad(self, data):
        return len(data[data[self.label_name] == 0])

    def _count_good(self, data):
        return len(data[data[self.label_name] == 1])

    def _bad_rate(self, data):
        return self._count_bad(data) * 1.0 / len(data)

    # 分数分布，可输入分数的处理函数，可以用来求beta sum
    def get_distribution(self, col_name=None, max_=1.0, min_=0.0, step_size=0.01, func=None):
        col_name = col_name if col_name is not None else self.score_name
        arr = np.arange(min_, max_, step=step_size)
        arr = arr.tolist()
        arr.append(max_)
        result = []
        data = self.data.copy()
        if func is not None and hasattr(func, '__call__'):
            data[col_name] = data[col_name].apply(func)
        for ind in range(1, len(arr)):
            col_data = data[col_name].between(arr[ind - 1], arr[ind]).values.sum()
            result.append(col_data)
        return result, arr

if __name__ == "__main__":
    pass    



# -*- coding: utf-8 -*-
"""
This file define a function which can

output a formatted data analysis csv file.

"""
import os
import time
import pandas as pd
from .Eval.CfMetric import CfMetric
from .Eval.Metric import Metric
import numpy as np
import sys
sys.dont_write_bytecode = True


def feat_split_label_cross(ofname, data, select_titles=None, missing_value=-1, group_list = None, group_column = None, 
                           month_list = None, month_column = None, label_column = None, good_label = 1):
    '''
    ofname(str)：结果导出路径
    select_titles(list)：进行分析的列名
    missing_value(int)：缺失值定义
    group_list(list-2d like [["cate_name",[cate_value]],["cate_name",[cate_value]]...])：分组列表
    group_column(string)：待分析的DataFrame中用于分组主键的列名
    month_list(list)：按月分组的列表
    month_column(str)：月份字段在DataFrame中的列名
    label_column(str)：待分析的DataFrame中用于标签的列名
    good_label(int)：好人的标签


    return a local file
    '''

    print('\n\n*** uni_variable analyzing... ***')
    print('output path: %s\n' % ofname)
    
    dat0 = data
    
    if isinstance(month_list,list):
        psi_title = ''
        avg_title = ',avg'+month_list[0]
        for i in range(1, len(month_list)):
            psi_title += ',psi' + month_list[i]
            avg_title += ',avg' + month_list[i]
    else:
        psi_title = ''
        avg_title = ''
        
    outfile = open(ofname,'w')
    outfile.write('id,group,feat,desc,bin,miss_rate,lowbound,upbound,average,' \
               + 'total,good,bad,user_rate,bad_rate,ks,iv'
               + psi_title + avg_title + '\n')
    ids = 0
    
    #如果没有传group_list默认进行全部的分析
    all_group = ['all',[]]
    if group_list:
        group_list.insert(0,all_group)
    else:
        group_list=[all_group]
        
    for g in group_list:
        k,v = g
        if k == 'all':
            dat = dat0
        else:
            assert group_column, "If you wanna analysis split by group, a group column name list should be given"
            dat = dat0[dat0[group_column].isin(v)]
        
        if select_titles is None:
            print ("Need input columns name that you want to analysis!")
            return 0
        print ("\n================= GROUP "+str(k)+" =================")
        # 整体
        n_tot = len(dat)
        n_god = len(dat[dat[label_column]==good_label])
        n_bad = n_tot - n_god
        for t in select_titles:
            print ("[" + str(t) + "]")
            # 缺失统计
            n_t_mis_god = len(dat[np.logical_and(dat[t] == missing_value, dat[label_column]==1)])
            n_t_mis_bad = len(dat[np.logical_and(dat[t] == missing_value, dat[label_column]==0)])
            n_mis  = n_t_mis_god + n_t_mis_bad
            mis_rate = 0.0 if n_tot==0 else n_mis*1.0/n_tot
            bad_rate = 0.0 if n_tot==0 else n_bad*100.0/n_tot
            
            dat_t = dat[dat[t]!=missing_value]
            vavg = dat_t[t].mean()
            vmin = dat_t[t].min()
            vmax = dat_t[t].max()
            # ks,iv,psi
            ks, iv = (-1, -1)
            if len(dat_t) > 1 and len(set(dat_t[label_column].values)) == 2:
                try:
                    # print('debug:\n', dat_t[label_column].values, dat_t[t].values)
                    met = CfMetric(y_true=dat_t[label_column].values, y_score=dat_t[t].values, seg=10)
                    ks = met.ks
                    iv = met.iv
                except IndexError:
                    pass
            
            psi_str = ''
            avg_str = ''
            if month_list and len(month_list)>=2:
                mt = Metric(dat_t, label_name=label_column, score_name=t)
                psi_on_given_splits = mt.get_psi_given_splits(month_list, split_name = month_column)
                tmp = dat_t[dat_t[month_column] == month_list[0]]
                if len(tmp) > 0:
                    avg_str +=  ',%g' % (tmp[t].mean())
                else:
                    avg_str += ','
                for i in range(1, len(month_list)):
                    if month_list[i] in psi_on_given_splits:
                        psi_str += ',%g' % psi_on_given_splits[month_list[i]]
                    else:
                        psi_str += ','
                    tmp = dat_t[dat_t[month_column] == month_list[i]]
                    if len(tmp) > 0:
                        avg_str +=  ',%g' % (tmp[t].mean())
                    else:
                        avg_str += ','
            else:
                psi_str += ','
                avg_str += ','
            
            ids += 1
            outfile.write('%d,%s,%s,,ALL,%.3f,%g,%g,%g,%d,%d,%d,,%.3f,%.2f,%.3f%s%s\n' % \
                       (ids, k, t, mis_rate, vmin, vmax, vavg, n_tot, n_god, 
                        n_bad, bad_rate, ks, iv, psi_str, avg_str))
            ids += 1
            bad_rate = 0.0 if n_mis==0 else n_t_mis_bad*100.0/n_mis
            outfile.write('%d,%s,%s,,MISS,,,,,%d,%d,%d,,%.3f\n' % \
                       (ids, k, t, n_mis, n_t_mis_god, n_t_mis_bad, bad_rate))
            v_set = set(dat_t[t].values)
            ## 分组分析，如果取值较少，则每个值分一组；否则用 qcut 十分位分析
            if len(v_set) < 4:
                i = 0
                for vv in v_set:
                    i += 1
                    n_t_v_god = len(dat_t[(dat_t[t] == vv) & (dat_t[label_column]==1)])
                    n_t_v_bad = len(dat_t[(dat_t[t] == vv) & (dat_t[label_column]==0)])
                    n_t_v  = n_t_v_god + n_t_v_bad
                    ids += 1
                    use_rate = 0.0 if n_tot==0.0 else n_t_v*100.0/n_tot
                    bad_rate = 0.0 if n_t_v==0.0 else n_t_v_bad*100.0/n_t_v
                    outfile.write('%d,%s,%s,,%d,,%g,,,%d,%d,%d,%.3f,%.3f\n' % \
                               (ids, k, t, i, vv, n_t_v, 
                                n_t_v_god, n_t_v_bad, use_rate, bad_rate))
            else:
                _, bins = pd.qcut(dat_t[t], 10, duplicates='drop', retbins=True)
                # print('bins', bins)
                for i in range(1,len(bins)):
                    tmp = dat_t[(dat_t[t]>bins[i-1]) & (dat_t[t]<=bins[i])]
                    vavg = tmp[t].mean()
                    n_t_i = len(tmp)
                    n_t_i_god = len(tmp[tmp[label_column]==1])
                    n_t_i_bad = n_t_i - n_t_i_god
                    ids += 1
                    use_rate = 0.0 if n_tot==0.0 else n_t_i*100.0/n_tot
                    bad_rate = 0.0 if n_t_i==0.0 else n_t_i_bad*100.0/n_t_i
                    outfile.write('%d,%s,%s,,%d,,%f,%f,%g,%d,%d,%d,%.3f,%.3f\n' % \
                               (ids, k, t, i, bins[i-1], bins[i], vavg, n_t_i, 
                                n_t_i_god, n_t_i_bad, use_rate, bad_rate))
            outfile.flush()
    outfile.close()
    
        
if __name__ == "__main__":
    prefix = time.strftime('%Y%m%d_%H%M',time.localtime(time.time()))
    outdir = './analysis/'
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    ofname = outdir + ('log.feat_split_label_cross.%s.csv' % prefix)

    data = 'load data to here'
    data = pd.read_csv('test.csv',sep=',',encoding='utf-8')
    
    select_titles = ['card_attr','flag_card_high_end','cna_score']
    group_list = [['new',[1]],['std',[2]]]
    
    month_list=['2018-11','2018-12']
    
    feat_split_label_cross(ofname, data, select_titles=select_titles, missing_value=-1, group_list = group_list, group_column = 'group', 
                           month_list = month_list, month_column = 'month', label_column = 'label', good_label = 1)
    



# coding=utf-8
from .CfMetric import CfMetric
from .PSI import calc_psi_list
#from Tools.ListTools import *
import numpy as np
import sys
sys.dont_write_bytecode = True

# 入参是array，不是pandas
class BaseMetric:
    def __init__(self, score):
        assert isinstance(score, (list, np.ndarray)), "support list_like score only (for now)"
        self.score = np.array(score)

    def get_ks(self, label, list_=False,  bad_rate=False):
        assert isinstance(label, (list, np.ndarray)), "support list_like score only (for now), current label type=" + str(type(label))
        assert len(self.score) == len(label), "label dimension is not matching.."
        label = np.array(label)
        re = CfMetric(y_true=label, y_score=self.score)
        if list_:
            print(re.ks_list)
        if bad_rate:
            tot_bad = sum(label == 0)
            tot_good = sum(label == 1)
            bad_rate = tot_bad * 1.0 / len(self.score)
            return re.ks, bad_rate, tot_bad, tot_good
        return re.ks

    def get_percentile(self, p):
        # assert is_number_list(p, threshold=100), 'p should be list of number, [0, 100]'
        return np.percentile(self.score, p)

    def min(self):
        return self.score.min()

    def max(self):
        return self.score.max()


# PSI
def get_psi(data, data2, bins=10, list_=False):
    cs = calc_psi_list(data, data2, nbins=bins)
    if list_:
        print(cs["psi_value"])
    return cs["psi_value"].sum()

if __name__ == "__main__":
    pass





